#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/controller.h>
#include <zephyr/bluetooth/gatt.h>
#include <bluetooth/gatt_dm.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/hci_vs.h>
#include <host/hci_core.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/logging/log.h>
#include <zephyr/settings/settings.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/sys/reboot.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <bluetooth/services/dfu_smp.h>
#include <zephyr/mgmt/mcumgr/transport/smp_bt.h>
#include <zephyr/mgmt/mcumgr/mgmt/mgmt.h>
#include <zephyr/mgmt/mcumgr/mgmt/callbacks.h>
//#include <bluetooth/scan.h>
//#include <zephyr/subsys/bluetooth/host/hci_core.h>

#include "ble_advertiser.h"

#include <zephyr/devicetree.h>
#include <zephyr/devicetree/fixed-partitions.h>
#include <zephyr/drivers/flash.h>
#include <zephyr/storage/flash_map.h>

#define LOG_MODULE_NAME ble_adv
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

#define TEMPERATURE_IDX 0
#define HUMIDITY_IDX 1

#define LOG_IP(x) ((uint8_t)(x / 100))
#define LOG_FP(x) ((uint8_t)(x - ((x / 100) * 100)))

#define SEC_TO_MS(x) (x * 1000)
#define MINUTES_TO_SEC(x) (x * 60)

#define MINUTES_1_IN_S 60
#define MINUTES_10_IN_MS 10 * SEC_TO_MS(MINUTES_1_IN_S)

#define HYPER_MODE_NC_PERIOD_MINUTES 30
#define ADV_HYPER_MSEC SEC_TO_MS(15)
#define ADV_NOT_HYPER_MSEC SEC_TO_MS(MINUTES_TO_SEC(5))
#define ADV_TIMEOUT_HSEC 100

#define DEVICE_NAME CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN (sizeof(DEVICE_NAME) - 1)
#define HASHSIZE (256 / 8)

#define MAX_DEVICE_ERRORS 2

#define ADVERTISE_CUSTOMDATA_SIZE (BT_UUID_SIZE_128 + sizeof(target_temp) + sizeof(inc_btn) + sizeof(dec_btn) + sizeof(connected_byte)+ sizeof(adv_numb) + sizeof(curr_adv_int) + sizeof(fw_hash))
#define AD_FW_HASH_OFFSET (BT_UUID_SIZE_128 + sizeof(target_temp) + sizeof(inc_btn) + sizeof(dec_btn) + sizeof(connected_byte)+ sizeof(adv_numb) + sizeof(curr_adv_int))

static uint8_t fw_hash[HASHSIZE];

/* ADV_MSEC is used as delay between scheduled adv work */
static uint32_t ADV_MSEC = 0;

static volatile uint8_t adv_numb = 0;
static bool target_sync = false;

/* Error list service */
#define BT_UUID_ERROR_LIST_SERVICE_VAL BT_UUID_128_ENCODE(0xac117609, 0x856c, 0x4bd7, 0xa78c, 0x03765174bc1b)
#define BT_UUID_ERROR_LIST_SERVICE BT_UUID_DECLARE_128(BT_UUID_ERROR_LIST_SERVICE_VAL)

/* software version */
#define BT_UUID_SW_VER_SERVICE_VAL BT_UUID_128_ENCODE(0x1e6e6e40, 0xfaf6, 0x4098, 0x90e7, 0x619658651741)
#define BT_UUID_SW_VER_SERVICE BT_UUID_DECLARE_128(BT_UUID_SW_VER_SERVICE_VAL)

/* custom data */
#define BT_UUID_CUSTOM_DATA_SERVICE_VAL BT_UUID_128_ENCODE(0xc90b9030, 0x0699, 0x4d89, 0x95a6, 0xde28e9f01e5c)
#define BT_UUID_CUSTOM_DATA_SERVICE BT_UUID_DECLARE_128(BT_UUID_CUSTOM_DATA_SERVICE_VAL)

static uint8_t errorlist_data[BT_UUID_SIZE_128 + sizeof(uint32_t) * MAX_DEVICE_ERRORS] = {BT_UUID_ERROR_LIST_SERVICE_VAL, 0x0};

static void start_advertising_coded(struct k_work* work);
static void adv_sent_cb(struct bt_le_ext_adv* adv,
                        struct bt_le_ext_adv_sent_info* info);
static void adv_conn_cb(struct bt_le_ext_adv* adv,
                        struct bt_le_ext_adv_connected_info* info);
static void adv_scanned_cb(struct bt_le_ext_adv* adv,
                           struct bt_le_ext_adv_scanned_info* info);
static void connected(struct bt_conn* conn, uint8_t conn_err);
static void disconnected(struct bt_conn* conn, uint8_t reason);

static void pairing_complete(struct bt_conn* conn, bool bonded);
static void pairing_failed(struct bt_conn* conn, enum bt_security_err reason);
void mtu_updated(struct bt_conn* conn, uint16_t tx, uint16_t rx);
static void timer_hyper_cb(struct k_timer* not_used);
static void timer_hyper_work_cb(struct k_work* work);
static void update_adv_msec(uint32_t ms);
static uint32_t get_adv_msec(void);

static K_WORK_DELAYABLE_DEFINE(start_advertising_worker, start_advertising_coded);
static K_TIMER_DEFINE(timer_hyper, timer_hyper_cb, NULL);
static K_WORK_DEFINE(timer_hyper_work, timer_hyper_work_cb);
K_SEM_DEFINE(adv_msec_sem, 1, 1);

static struct bt_le_ext_adv* adv;
static struct bt_conn* def_con;
static const uint32_t ADV_OPTS = (BT_LE_ADV_OPT_CONNECTABLE | BT_LE_ADV_OPT_EXT_ADV |
                                  BT_LE_ADV_OPT_CODED | BT_LE_ADV_OPT_USE_IDENTITY |
                                  BT_LE_ADV_OPT_ONE_TIME);

static const uint32_t ADV_DURING_CON_OPTS = (BT_LE_ADV_OPT_EXT_ADV | BT_LE_ADV_OPT_CODED |
                                             BT_LE_ADV_OPT_USE_IDENTITY);

static struct bt_le_adv_param adv_param = BT_LE_ADV_PARAM_INIT(ADV_OPTS,
                                                               BT_GAP_ADV_FAST_INT_MIN_2,
                                                               BT_GAP_ADV_FAST_INT_MAX_2,
                                                               NULL);

static struct bt_le_adv_param adv_conn_param = BT_LE_ADV_PARAM_INIT(ADV_DURING_CON_OPTS,
                                                                    BT_GAP_ADV_FAST_INT_MIN_2,
                                                                    BT_GAP_ADV_FAST_INT_MAX_2,
                                                                    NULL);

static struct bt_le_ext_adv_cb adv_cbs = {
    .sent = adv_sent_cb,
    .connected = adv_conn_cb,
    .scanned = adv_scanned_cb,
};

static struct bt_le_ext_adv_start_param ext_adv_start_param = {
    /* N * 10 ms */
    .timeout = ADV_TIMEOUT_HSEC,
};

static struct bt_conn_auth_info_cb conn_auth_info_callbacks = {
    .pairing_complete = pairing_complete,
    .pairing_failed = pairing_failed,
    .bond_deleted = NULL};

static struct bt_gatt_cb gatt_cbs = {
    .att_mtu_updated = mtu_updated,
};

static struct bt_conn_cb conn_cbs = {
    .connected = connected,
    .disconnected = disconnected,
};

typedef struct
{
    int status;
    uint8_t reason;
    const uint8_t* data;
    size_t data_len;
} ble_return_data_t;

static void ble_cb(ble_return_data_t data);
struct cb_context
{
    struct k_work work;
    ble_return_data_t data;
};

atomic_t ble_adv_active = ATOMIC_INIT(0);
atomic_t ble_scanner_active = ATOMIC_INIT(0);
atomic_t ble_connection_active = ATOMIC_INIT(0);

static bool ble_is_advertising(void)
{
    int is_adv = atomic_get(&ble_adv_active);

    return (1 == is_adv);
}

static bool ble_is_scanning(void)
{
    int is_scan = atomic_get(&ble_scanner_active);

    return (1 == is_scan);
}

static bool ble_is_connected(void)
{
    int is_connect = atomic_get(&ble_connection_active);

    return (1 == is_connect);
}

bool ble_is_active(void)
{
    bool is_adv = ble_is_advertising();
    bool is_scan = ble_is_scanning();
    bool is_connect = ble_is_connected();

    return (is_adv || is_scan || is_connect);
}

static void cb_handler(struct k_work* item)
{
    struct cb_context* ctx = CONTAINER_OF(item, struct cb_context, work);

    ble_cb(ctx->data);

    k_free((void*)(ctx->data.data));
    k_free(ctx);
}

static void submit_cb(int status, uint8_t reason, const uint8_t* data, size_t data_len)
{
    struct cb_context* ctx = k_malloc(sizeof(struct cb_context));

    if (ctx != NULL) {
        if (data != NULL) {
            ctx->data.data = k_malloc(data_len);

            if (ctx->data.data != NULL) {
                memcpy((void*)(ctx->data.data), data, data_len);
                ctx->data.data_len = data_len;
            }
        } else {
            ctx->data.data = NULL;
            ctx->data.data_len = 0;
        }

        ctx->data.status = status;

        k_work_init(&(ctx->work), cb_handler);
        k_work_submit(&(ctx->work));
    }
}

void mtu_updated(struct bt_conn* conn, uint16_t tx, uint16_t rx)
{
    LOG_INF("MTU updated to tx: %d and rx: %d", tx, rx);
}

static void pairing_complete(struct bt_conn* conn, bool bonded)
{
    char addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    bt_security_t security = bt_conn_get_security(conn);
    uint8_t key_size = bt_conn_enc_key_size(conn);

    LOG_INF("\n----> Paired to %s with security level %d and key size: %d bonded: %d\n", addr, security, key_size, bonded);
}

static void pairing_failed(struct bt_conn* conn, enum bt_security_err reason)
{
    char addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    LOG_ERR("Pairing failed for conn: %s, reason %d, disconnection", addr, reason);

    bt_conn_disconnect(conn, BT_HCI_ERR_PAIRING_NOT_SUPPORTED);

    if (def_con != NULL) {
        bt_conn_unref(def_con);
        def_con = NULL;
    } else {
        LOG_ERR("Pairing failed but had no connection object");
    }
}

static void connected(struct bt_conn* conn, uint8_t conn_err)
{
    int err;
    struct bt_conn_info info;
    char addr[BT_ADDR_LE_STR_LEN];

    atomic_set(&ble_connection_active, 1);

    def_con = bt_conn_ref(conn);
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    if (bt_addr_le_is_bonded(BT_ID_DEFAULT, bt_conn_get_dst(conn))) {
        struct bt_conn_le_phy_param phy_param = {
            .options = BT_CONN_LE_PHY_OPT_CODED_S8,
            .pref_rx_phy = BT_GAP_LE_PHY_CODED,
            .pref_tx_phy = BT_GAP_LE_PHY_CODED};

        if (bt_conn_le_phy_update(conn, &phy_param) != 0) {
            LOG_ERR("Failed to update PHY params");
        }
    }

    if (conn_err) {
        LOG_ERR("Connection failed (err %d)\n", conn_err);
        submit_cb(BT_CONNECTION_ERROR, conn_err, NULL, 0);
        atomic_set(&ble_connection_active, 0);
        return;
    }

    err = bt_conn_get_info(conn, &info);
    if (err) {
        LOG_ERR("Failed to get connection info (err %d)\n", err);
    } else {
        const struct bt_conn_le_phy_info* phy_info;
        phy_info = info.le.phy;

        LOG_INF("Connected: %s, tx_phy %u, rx_phy %u",
                addr, phy_info->tx_phy, phy_info->rx_phy);
    }
}

static void disconnected(struct bt_conn* conn, uint8_t reason)
{
    if (reason == BT_HCI_ERR_REMOTE_USER_TERM_CONN) {
        LOG_INF("Disconnected. Normal disconnect by central.");
    } else if (reason == BT_HCI_ERR_PAIRING_NOT_SUPPORTED) {
        LOG_ERR("Disconnected due to paring not supported");
    } else {
        LOG_ERR("Disconnected (reason 0x%02x)", reason);
    }

    if (def_con == conn) {
        bt_conn_unref(def_con);
        def_con = NULL;
    }
    atomic_set(&ble_connection_active, 0);
}

void ble_advertiser_schedule_adv(k_timeout_t delay)
{
    k_work_cancel_delayable(&start_advertising_worker);

    k_work_reschedule(&start_advertising_worker, delay);
}

static void adv_sent_cb(struct bt_le_ext_adv* adv, struct bt_le_ext_adv_sent_info* info)
{
    atomic_set(&ble_adv_active, 0);
    uint32_t temp = get_adv_msec();
    k_work_reschedule(&start_advertising_worker, K_MSEC(temp));
    LOG_INF("adv timeout reached (adv stopped). adv start rescheduled in: %d ms", temp);
}

static void adv_conn_cb(struct bt_le_ext_adv* adv, struct bt_le_ext_adv_connected_info* info)
{
    def_con = info->conn;
}

static void adv_scanned_cb(struct bt_le_ext_adv* adv, struct bt_le_ext_adv_scanned_info* info)
{
    char scan_addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(info->addr, scan_addr, BT_ADDR_LE_STR_LEN);

    LOG_INF("adv_scanned_cb called (due to AUX_SCAN_REQ) by %s", scan_addr);
}

static void start_advertising_coded(struct k_work* work)
{
    int err;
    uint8_t bat_lev = 0;
    uint16_t sd[2] = {10,10};  //fake data
    uint16_t target_temp =0x1122; //fake
    uint32_t nbr_of_errors = 0;
    uint8_t sw_ver_data[] = {BT_UUID_SW_VER_SERVICE_VAL, VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH};

    if ((err = bt_le_ext_adv_stop(adv)) < 0) {
        LOG_ERR("bt le ext adv stop error: %d", err);
    }

    uint8_t inc_btn = 0;
    uint8_t dec_btn = 0;
    uint8_t new_target[2] = {0};

    new_target[0] = LOG_IP(target_temp);
    new_target[1] = LOG_FP(target_temp);

    if (inc_btn) {
        new_target[0] += inc_btn;
    }

    if (dec_btn) {
        new_target[0] -= dec_btn;
    }

    LOG_INF("set_target(%hhu.%02hhu)", new_target[0], new_target[1]);

    target_temp = (new_target[0] * 100) + new_target[1];

    bat_lev = 25; // fake

    bool is_connected = true; //fake
    uint8_t connected_byte = is_connected ? 0xff : 0x00;

    LOG_INF("target_temp: %hhu.%hhuC (%hu) inc_btn: %hhu, dec_btn: %hhu",
            LOG_IP(target_temp), LOG_FP(target_temp), target_temp, inc_btn, dec_btn);

    LOG_INF(
        "Updated adv data; temp: %hhu.%02hhu, humid: %hhu.%02hhu, "
        "target temp: %hu, battery level: %hhu%% button inc: %04hu dec: %04hu",
        LOG_IP(sd[TEMPERATURE_IDX]), LOG_FP(sd[TEMPERATURE_IDX]),
        LOG_IP(sd[HUMIDITY_IDX]), LOG_FP(sd[HUMIDITY_IDX]),
        target_temp, bat_lev, inc_btn, dec_btn);

    adv_numb++;

    /* ADV_MSEC is in ms. Advertise in seconds */
    uint16_t curr_adv_int = (uint16_t) (get_adv_msec() / 1000);

    if (!target_sync) {
        inc_btn = 0;
        dec_btn = 0;
    }

    uint8_t custom_data[ADVERTISE_CUSTOMDATA_SIZE] = {BT_UUID_CUSTOM_DATA_SERVICE_VAL, BT_UUID_16_ENCODE(target_temp), inc_btn, dec_btn, connected_byte,
                                                      adv_numb, BT_UUID_16_ENCODE(curr_adv_int)};

    memcpy(&(custom_data[AD_FW_HASH_OFFSET]), &(fw_hash[0]), sizeof(fw_hash));

    //LOG_HEXDUMP_INF(custom_data, ADVERTISE_CUSTOMDATA_SIZE, "Custom advertise data");

    uint8_t smpdata[BT_UUID_SIZE_128] = {BT_UUID_DFU_SMP_SERVICE_VAL};

    struct bt_data ad[] = {
        BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_LIMITED | BT_LE_AD_NO_BREDR)),
        BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
        BT_DATA_BYTES(BT_DATA_SVC_DATA16, BT_UUID_16_ENCODE(BT_UUID_BAS_VAL), BT_UUID_16_ENCODE(bat_lev)),
        BT_DATA_BYTES(BT_DATA_SVC_DATA16, BT_UUID_16_ENCODE(BT_UUID_TEMPERATURE_VAL), BT_UUID_16_ENCODE(sd[TEMPERATURE_IDX])),
        BT_DATA_BYTES(BT_DATA_SVC_DATA16, BT_UUID_16_ENCODE(BT_UUID_HUMIDITY_VAL), BT_UUID_16_ENCODE(sd[HUMIDITY_IDX])),
        BT_DATA(BT_DATA_SVC_DATA128, custom_data, sizeof(custom_data)),
        BT_DATA(BT_DATA_SVC_DATA128, smpdata, sizeof(smpdata)),
        BT_DATA(BT_DATA_SVC_DATA128, errorlist_data, BT_UUID_SIZE_128 + sizeof(uint32_t) * nbr_of_errors),
        BT_DATA(BT_DATA_SVC_DATA128, sw_ver_data, sizeof(sw_ver_data))
    };

    /* Continue advertising while connected, but NCONN param */
    if (def_con) {
        if ((err = bt_le_ext_adv_update_param(adv, &adv_conn_param)) < 0) {
            LOG_ERR("Could not update NCONN adv params (err %d)", err);
        }
    } else {
        if ((err = bt_le_ext_adv_update_param(adv, &adv_param)) < 0) {
            LOG_ERR("Could not update CONN adv params (err %d)", err);
        }
    }

    atomic_set(&ble_adv_active, 1);

    if ((err = bt_le_ext_adv_set_data(adv, ad, ARRAY_SIZE(ad), NULL, 0))) {
        LOG_ERR("Failed to set advertising data (err %d)", err);
    }

    if ((err = bt_le_ext_adv_start(adv, &ext_adv_start_param))) {
        LOG_ERR("Failed to start advertising set (err %d)", err);
    }

    LOG_INF("Advertising started\n");
}

static void reset_bond(const struct bt_bond_info* info, void* user_data)
{
    bt_unpair(BT_ID_DEFAULT, &info->addr);
}

static void reset_bond_information(void)
{
    bt_foreach_bond(BT_ID_DEFAULT, reset_bond, NULL);
}

void release_bonded_devices(void)
{
    reset_bond_information();
}

static void handle_adv_int_update(const uint8_t* data)
{
    /* We receive in seconds, but need to set in ms */

    uint32_t new_sec = (data[0] << 8) | data[1];
    LOG_ERR("updated adv int: %u s", new_sec);

    if (new_sec > MINUTES_TO_SEC(10)) {
        LOG_WRN("New adv int too ..... high. Setting max val (10 min)");
        update_adv_msec(MINUTES_10_IN_MS);
        return;
    }

    update_adv_msec(SEC_TO_MS(((uint32_t)new_sec)));
    uint32_t temp = get_adv_msec();
    LOG_DBG("New ADV_MSEC: %d ms", temp);

    k_timer_stop(&timer_hyper);
    ble_advertiser_schedule_adv(K_MSEC(temp));
}

static void remote_error_led_indication(const uint8_t* data, const size_t len)
{
    bool signal = false;

    if (data[0] == 1) {
        signal = true;
    }
}

static void set_received_errors(const uint8_t* data, const size_t len)
{

}

static void update_adv_msec(uint32_t ms)
{
    if ((k_sem_take(&adv_msec_sem, K_NO_WAIT)) == 0) {
        ADV_MSEC = ms;
        k_sem_give(&adv_msec_sem);
    } 
}

static uint32_t get_adv_msec(void)
{
    uint32_t temp;
    if ((k_sem_take(&adv_msec_sem, K_NO_WAIT)) == 0) {
        temp = ADV_MSEC;
        k_sem_give(&adv_msec_sem);
    } else {
        LOG_ERR("Could not read ADV_MSEC. Setting 15 s");
        temp = SEC_TO_MS(15);
    }

    return temp;
}


static void ble_cb(ble_return_data_t data)
{
    switch (data.status) {
        case BT_CONNECTED:
            LOG_INF("\"Connected\" to gateway");
            break;
        case BT_DATA_RECEIVED:
            target_sync = true;
            LOG_INF("---> Received new target %hhu.%hhuC", data.data[0], data.data[1]);

            break;
        case BT_CONNECTION_ERROR:
            LOG_ERR("Disconnected because of reason %hhu", data.reason);
            break;
        case BT_FACTORY_RESET:
            LOG_INF("Factory reset requested");
            break;
        case BT_ADV_INT_RECEIVED:
            LOG_INF("Received new advertising interval");
            handle_adv_int_update(data.data);
            break;
        case BT_BLEBLE_ADDR_RECEIVED:
            LOG_INF("Received BLEBLE client addr");

            break;
        case BT_BLEBLE_CLIENT_SS_RX:
            /**
             * INFO:
             * We are now out of sync with bleblehost, since we've been connected to nrf-central.
             *    - Let it be and have bleblehost scan actively for < 5 minutes.
             *
             * If we want to set NCONN after nrf-central has informed us that we're a blebleclient,
             * we could set:
             * adv_param = adv_conn_param
             * While this would inhibit further connections (and therefore remove the risk of
             * nrf-central connecting to us and changing our adv int), we will be in hyper mode for
             * 30 minutes every time an NFC-blipp is done.
             */
            break;
        case BT_ERRORLIST_RECEIVED:
            LOG_INF("Received errorlist");
            set_received_errors(data.data, data.data_len);
            break;
        case BT_REMOTE_ERROR_LED_INIDCATION:
            LOG_INF("Received remote led error signal change");
            remote_error_led_indication(data.data, data.data_len);
            break;
        default:
            LOG_INF("Not a know status code: %d", data.status);
            break;
    }
}

void timer_hyper_cb(struct k_timer* not_used)
{
    k_work_submit(&timer_hyper_work);
}

void timer_hyper_work_cb(struct k_work* work)
{
    LOG_ERR("Hyper mode timer expired");
    update_adv_msec(ADV_NOT_HYPER_MSEC);

    /* Start advertising once directly to advertise with new intervall info, any quiet period is handled by the next re-scheduling in adv_sent_cb */
    ble_advertiser_schedule_adv(K_NO_WAIT);
}

void ble_nfc_blipp_notify(void)
{
    LOG_INF("ble_nfc_blipp_notify: Starting hyper adv");

    update_adv_msec(ADV_HYPER_MSEC);

    k_timer_start(&timer_hyper, K_MINUTES(HYPER_MODE_NC_PERIOD_MINUTES), K_FOREVER);
    ble_advertiser_schedule_adv(K_NO_WAIT);
}

int ble_adv_init(void)
{
    int err;

    if ((err = bt_enable(NULL))) {
        LOG_ERR("Bluetooth init failed (err %d)", err);
    }

    LOG_INF("Bluetooth initialized");

    if ((err = settings_load())) {
        LOG_ERR("settings_load failed (err %d)", err);
    }

    if ((err = bt_le_ext_adv_create(&adv_param, &adv_cbs, &adv))) {
        LOG_ERR("Failed to create advertiser set (err %d)", err);
    }

    smp_bt_register();

    bt_gatt_cb_register(&gatt_cbs);
    bt_conn_cb_register(&conn_cbs);
    bt_conn_auth_info_cb_register(&conn_auth_info_callbacks);

    if (true) {
        ble_nfc_blipp_notify();
    }

    return err;
}
