#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <host/hci_core.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/services/dfu_smp.h>
#include <zephyr/bluetooth/bluetooth.h>

#include <zephyr/bluetooth/controller.h>
#include <zephyr/bluetooth/hci_vs.h>
#include <bluetooth/scan.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/logging/log.h>

#if defined(CONFIG_BT_SETTINGS)
#include <zephyr/settings/settings.h>
#endif /* CONFIG_BT_SETTINGS */

#include "sensor_dfu_handler.h"
#include "peer_utils.h"
#include "ble_scanner_new.h"

#define LOG_MODULE_NAME ble_new
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

#define BLE_SCAN_TYPE                                     BT_LE_SCAN_TYPE_ACTIVE
#define BLE_SCAN_INTERVAL                              BT_GAP_SCAN_FAST_INTERVAL
#define BLE_SCAN_WINDOW                                  BT_GAP_SCAN_FAST_WINDOW
#define BLE_SCAN_OPTS             (BT_LE_SCAN_OPT_CODED | BT_LE_SCAN_OPT_NO_1M )

#define BLE_CONN_OPS               (BT_CONN_LE_OPT_CODED | BT_CONN_LE_OPT_NO_1M)
#define BLE_CONN_INT                                   BT_GAP_SCAN_FAST_INTERVAL
#define BLE_CONN_WIN                                     BT_GAP_SCAN_FAST_WINDOW

#define MAX_NEW_DATA_QUEUE_ATTEMPTS 2

#define NEW_DATA_MAX_MSGS 120
#define NEW_DATA_THREAD_STACK_SIZE 4096
#define NEW_DATA_THREAD_PRIORITY 12

static int scan_start(void);
static int scan_stop(void);
static void scan_timeout(void);
static void scan_recv(const struct bt_le_scan_recv_info* info,
                      struct net_buf_simple* buf);

static void connected(struct bt_conn *conn, uint8_t conn_err);
static void disconnected(struct bt_conn *conn, uint8_t reason);
static void pairing_failed(struct bt_conn *conn, enum bt_security_err reason);
static void pairing_complete(struct bt_conn *conn, bool bonded);
static void discover_completed(struct bt_gatt_dm *dm, void *ctx);
static void discover_service_not_found(struct bt_conn *conn, void *ctx);
static void discover_error_found(struct bt_conn *conn, int err, void *ctx);
static void security_changed(struct bt_conn* conn, bt_security_t level, 
                             enum bt_security_err err);
static void phy_updated(struct bt_conn* conn,
                        struct bt_conn_le_phy_info* param);
static void dfu_smp_error_cb(struct bt_dfu_smp* dfu_smp, int err);
static void peer_gatt_thread(void* u1, void* u2, void* u3);
static void new_data_thread(void *, void *, void *);

K_MSGQ_DEFINE(peer_gatt_msgq, sizeof(int), 10, 4);
K_MSGQ_DEFINE(new_data_msgq, sizeof(adv_le_user_data_t), NEW_DATA_MAX_MSGS, 2);
K_THREAD_DEFINE(peer_gatt_thread_id, 2048, peer_gatt_thread, NULL, NULL, NULL, 5, 0, 0);
K_THREAD_DEFINE(new_data_tid, NEW_DATA_THREAD_STACK_SIZE, new_data_thread, NULL, NULL, NULL, NEW_DATA_THREAD_PRIORITY, 0, 0);

#define PEER_QUEUE_STACK_SIZE 4096
#define PEER_QUEUE_PRIORITY 14

K_THREAD_STACK_DEFINE(peer_queue_stack_area, PEER_QUEUE_STACK_SIZE);

static struct k_work_q peer_work_queue;

static struct k_work_queue_config peer_work_queue_config ={.name = "peer_work_qeueu", .no_yield = false};
struct k_work_delayable discover_work;

static atomic_t ota_peer = ATOMIC_INIT(-1);

static struct bt_conn_le_create_param create_params = {
    .options        = BLE_CONN_OPS,
    .interval       = BLE_CONN_INT,
    .window         = BLE_CONN_WIN,
    .interval_coded = 0,
    .window_coded   = 0,
    .timeout        = 0,
};

static struct bt_le_conn_param conn_params = {
    .interval_min = BT_GAP_INIT_CONN_INT_MIN,
    .interval_max = BT_GAP_INIT_CONN_INT_MAX,
    .latency      = 1,
    .timeout      = 400,
};

static struct bt_le_scan_cb scan_le_cb = {
    .recv = scan_recv,
    .timeout = scan_timeout,
};

static struct bt_le_scan_param scan_param = {
    .type     = BLE_SCAN_TYPE,
    .interval = BLE_SCAN_INTERVAL,
    .window   = BLE_SCAN_WINDOW,
    .options  = BLE_SCAN_OPTS
};
static struct bt_conn_le_phy_param phy_param = {
    .options     = BT_CONN_LE_PHY_OPT_CODED_S8,
    .pref_rx_phy = BT_GAP_LE_PHY_CODED,
    .pref_tx_phy = BT_GAP_LE_PHY_CODED,
};
static struct bt_conn_auth_info_cb conn_auth_info_callbacks = {
	.pairing_complete = pairing_complete,
	.pairing_failed   = pairing_failed
};
static struct bt_conn_cb conn_cbs = {
    .connected        = connected,
    .disconnected     = disconnected,
    .security_changed = security_changed,
    .le_phy_updated   = phy_updated,
};

static struct bt_gatt_dm_cb discover_cb = {
    .completed         = discover_completed,
    .service_not_found = discover_service_not_found,
    .error_found       = discover_error_found,
};

static void ota_cleanup(int idx)
{
    //k_sem_give(get_peer_send_to_ota_sem(idx));
}

static void connection_cleanup(int peer_index)
{
    /* this will eventuelly end in ota_cb where ota_peer is cleared */
    ota_abort(peer_index);

    /* just set back all semaphore to avoid it being locked on next connect */
    k_sem_give(get_peer_send_to_ota_sem(peer_index));
}

static void ota_cb(int peer_index, ota_status_t status)
{
    int err;
    struct bt_conn* conn = get_client_conn();

    if (conn != NULL) {
        if ((err = bt_conn_disconnect(conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN)) != 0) {
            LOG_ERR("Failed to disconnet (err: %d)", err);
        } else {
            LOG_INF("Disconnecting");
        }
    } else {
        LOG_INF("ota_cb conn null");
    }

    atomic_set(&ota_peer, -1);

    LOG_INF("Ota completed status %d", status);

    ota_cleanup(peer_index);
}

static struct bt_dfu_smp_init_params dfu_init_params = {
    .error_cb = dfu_smp_error_cb,
};

static void dfu_smp_error_cb(struct bt_dfu_smp* dfu_smp, int err)
{
    LOG_ERR("dfu smp error %d", err);
}

static void phy_updated(struct bt_conn* conn, struct bt_conn_le_phy_info* param)
{
    LOG_DBG("Phy param update tx_phy %u, rx_phy %u", param->tx_phy, param->rx_phy);
}

static void pairing_complete(struct bt_conn *conn, bool bonded)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    if (bonded) {
        bt_security_t sec = bt_conn_get_security(conn);
        uint8_t key_size = bt_conn_enc_key_size(conn);
        LOG_DBG("Pairing completed: %s. Sec level: %u, key size: %u", addr, sec, key_size);
    }
}

static void pairing_failed(struct bt_conn *conn, enum bt_security_err reason)
{
    char addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    LOG_ERR("Pairing failed conn: %s, reason %d", addr, reason);
}

static void discover_completed(struct bt_gatt_dm *dm, void *ctx)
{
    int err;
    int idx;
    const struct bt_gatt_dm_attr* gatt_service_attr = NULL;
    const struct bt_gatt_service_val* gatt_service = NULL;
    struct bt_conn *conn = bt_gatt_dm_conn_get(dm);

    bt_gatt_dm_data_print(dm);

    gatt_service_attr = bt_gatt_dm_service_get(dm);
    gatt_service = bt_gatt_dm_attr_service_val(gatt_service_attr);

    if ((idx = get_peers_idx_from_bt_addr_le_t(bt_conn_get_dst(conn))) < 0) {
        LOG_ERR("peers[idx] out of bounds (err %d)", idx);
        return;
    }

    if (bt_uuid_cmp(gatt_service->uuid, BT_UUID_DFU_SMP_SERVICE) == 0) {

        if ((err = bt_dfu_smp_handles_assign(dm, get_peer_dfu_smp(0)))) {
            LOG_ERR("Failed to init dfu smp client (err %d)", err);
        }
    }

    if ((err = bt_gatt_dm_data_release(dm))) {
        LOG_ERR("Could not release the discovery data (err %d)", err);
    }

    bt_gatt_dm_continue(dm, NULL);
}

static void peer_gatt_thread(void* u1, void* u2, void* u3)
{
    int err;
    int avoid_disconnect;
    int idx;

    while (1) {
        if (k_msgq_get(&peer_gatt_msgq, &idx, K_FOREVER) == 0) {

            avoid_disconnect = false;

            if (atomic_cas(&ota_peer, -1, 0)) {
                if (k_sem_take(get_peer_send_to_ota_sem(idx), K_NO_WAIT) == 0) {
                    LOG_INF("Sending to ota thread");

                    peer_to_ota_thread(idx, ota_cb);
                    avoid_disconnect = true;

                } else {
                    LOG_ERR("Failed to get send_to_ota_sem");
                    ota_cleanup(idx);
                }
            } else {
                LOG_INF("Ota already running for another device, discard");
                ota_cleanup(idx);
            }

            if (!avoid_disconnect) {
                LOG_DBG("Attempting to disconnect in peer gatt thread");

                if ((err = bt_conn_disconnect(get_client_conn(), BT_HCI_ERR_REMOTE_USER_TERM_CONN)) != 0) {
                    LOG_ERR("Failed to disconnect (err: %d)", err);
                } else {
                    LOG_DBG("Disconnecting");
                }
            } else {
                LOG_DBG("Avoiding disconnect due to ongoing DFU");
            }
        }
    }
}

static void discover_service_not_found(struct bt_conn *conn, void *ctx)
{
    int err;
    int idx;

    if ((idx = get_peers_idx_from_bt_addr_le_t(bt_conn_get_dst(conn))) < 0) {
        LOG_ERR("peers[idx] out of bounds (err %d)", idx);
        return;
    }

    LOG_DBG("in discover_service_not_found. adding to peer gatt msgq");

    if ((err = k_msgq_put(&peer_gatt_msgq, &idx, K_NO_WAIT)) != 0) {
        LOG_ERR("Could not put peer in peer gatt msgq. err: %d", err);

        if ((err = bt_conn_disconnect(get_client_conn(), BT_HCI_ERR_REMOTE_USER_TERM_CONN)) != 0) {
            LOG_ERR("Failed to disconnect (err: %d)", err);
        } else {
            LOG_DBG("Disconnecting");
        }
    }
}

static void discover_error_found(struct bt_conn *conn, int err, void *ctx)
{
    LOG_ERR("Discovery error found (err: %d)", err);
}

static void exchange_func(struct bt_conn* conn, uint8_t err,
                         struct bt_gatt_exchange_params* params)
{
    LOG_DBG("MTU exchange %s", err == 0 ? "successful" : "failed");
    LOG_DBG("Current MTU: %u", bt_gatt_get_mtu(conn));
}

static struct bt_gatt_exchange_params exchange_params = {
    .func = exchange_func,
};

static void discover_work_handler(struct k_work* work)
{
    int err;
    char addr[BT_ADDR_LE_STR_LEN];

    if ((err = bt_gatt_exchange_mtu(get_client_conn(), &exchange_params))) {
        LOG_ERR("MTU exchange failed");
    } else {
        LOG_DBG("MTU exchange pending");
    }

    
    bt_addr_le_to_str(bt_conn_get_dst(get_client_conn()), addr, sizeof(addr));
    if ((err = bt_gatt_dm_start(get_client_conn(), NULL, &discover_cb, NULL))) {
        LOG_ERR("Failed to start discovery (err %d)", err);

        if ((err = bt_conn_disconnect(get_client_conn(), BT_HCI_ERR_REMOTE_USER_TERM_CONN)) != 0) {
            LOG_ERR("Failed to disconnet (err: %d)", err);
        } else {
            LOG_DBG("Disconnecting");
        }
    }
}

static void discovery_init(int peer_index)
{
    k_work_init_delayable(&(discover_work), discover_work_handler);
    k_work_schedule_for_queue(&peer_work_queue, &discover_work, K_NO_WAIT);
}

static void security_changed(struct bt_conn* conn, bt_security_t level, enum bt_security_err err)
{
	int idx;
	char dst[BT_ADDR_LE_STR_LEN];
	bt_addr_le_to_str(bt_conn_get_dst(conn), dst, sizeof(dst));

    if ((idx = get_peers_idx_from_bt_addr_le_t(bt_conn_get_dst(conn))) < 0) {
        LOG_ERR("idx error (idx: %d)", idx);
        return;
    }

    bool is_bonded = bt_addr_le_is_bonded(BT_ID_DEFAULT, bt_conn_get_dst(conn));

    if (!err && is_bonded) {
        LOG_DBG("Security changed: %s level %u", dst, level);

        discovery_init(idx);

    } else {
        int error;

        LOG_ERR("Security failed: %s level %u (err %d)", dst, level, err);

        if (err == BT_SECURITY_ERR_UNSPECIFIED) {
            LOG_WRN("Peripheral could be bonded with other device.");
            LOG_WRN("Try to factory reset the peripheral by long back button press");
        }

        if ((error = bt_unpair(BT_ID_DEFAULT, bt_conn_get_dst(conn))) < 0) {
            LOG_ERR("unpair error: %d", error);
        }

        if ((error = bt_conn_disconnect(conn, BT_HCI_ERR_PAIRING_NOT_SUPPORTED)) != 0) {
            LOG_ERR("Failed to disconnect after pairing + bonding (err: %d error %d)", err, error);

            /* We end up here when peripheral is re-flashed, but central is bonded with peripheral. */

        } else {
            LOG_WRN("Disconnecting due to security fail");
        }

        LOG_WRN("-> trigger reconnect event since security failed");

    }
}

static void connected(struct bt_conn *conn, uint8_t conn_err)
{
	int idx;
    int err;
	char dst[BT_ADDR_LE_STR_LEN];
	bt_addr_le_to_str(bt_conn_get_dst(conn), dst, sizeof(dst));

    (void)scan_start();

    if ((idx = get_peers_idx_from_bt_addr_le_t(bt_conn_get_dst(conn))) < 0) {
        LOG_ERR("idx error (idx: %d)", idx);
        return;
    }

    if (conn_err) {
        LOG_ERR("Failed to connect to %s, (err %u)", dst, conn_err);

        connection_cleanup(idx);

        if (get_client_conn() == conn) {
            bt_conn_unref(get_client_conn());
            set_client_conn(NULL);
            LOG_INF("Connection unreferenced");
        }

        return;
    }

    LOG_INF("Connected to %s rssi %d", dst, get_peer_rssi());
    bool is_bonded = bt_addr_le_is_bonded(BT_ID_DEFAULT, bt_conn_get_dst(conn));
    LOG_DBG("Bonded: %s", is_bonded ? "true" : "false");

    LOG_INF("Updating PHY params");
    if (bt_conn_le_phy_update(conn, &phy_param) != 0) {
        LOG_ERR("Failed to update PHY params");

        if ((err = bt_conn_disconnect(conn, BT_HCI_ERR_AUTH_FAIL)) != 0) {
            LOG_ERR("Failed to disconnect after update PHY params error. err: %d", err);
        } else {
            LOG_WRN("Disconnecting due to update PHY params");
        }
        return;
    }

    LOG_INF("Setting security L2");
    if (bt_conn_set_security(conn, BT_SECURITY_L2) != 0) {
        LOG_ERR("Could not set security");

        if ((err = bt_conn_disconnect(conn, BT_HCI_ERR_AUTH_FAIL)) != 0) {
            LOG_ERR("Failed to disconnect after set security error. err: %d", err);
        } else {
            LOG_WRN("Disconnecting due to set security fail");
        }
        return;
    }
}

/* allow one connection + one connection where OTA is running.
   and yes the name and logic is strange */
static bool connect_nok(void)
{
    if (get_client_conn() != NULL) {
            return true;
    }

    return false;
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	int idx;
	char dst[BT_ADDR_LE_STR_LEN];
	bt_addr_le_to_str(bt_conn_get_dst(conn), dst, sizeof(dst));

    if ((idx = get_peers_idx_from_bt_addr_le_t(bt_conn_get_dst(conn))) < 0) {
        LOG_ERR("peer idx error (%d)", idx);
    }

    set_peer_offline(0);

    if (reason == BT_HCI_ERR_LOCALHOST_TERM_CONN) {
        LOG_INF("Disconnected. Normal disconnect by central");
    } else {
        LOG_ERR("Disconnected: %s (reason 0x%02x)", dst, reason);
    }

    connection_cleanup(idx);

    if (get_client_conn() == conn) {
        bt_conn_unref(get_client_conn());
        set_client_conn(NULL);
        LOG_DBG("bt conn unref called on peers[%d].def_conn", idx);

    } else{
        LOG_ERR("peers[idx].def_conn != conn");
    }
}

bool parse_adv_data_cb(struct bt_data* data, void* user_data)
{
    switch (data->type) {
        case BT_DATA_FLAGS: {
            break;
        }

        case BT_DATA_NAME_COMPLETE: {
            break;
        }

        case BT_DATA_SVC_DATA16: {

            break;
        }

        case BT_DATA_UUID16_ALL: {
            break;
        }

        case BT_DATA_UUID128_ALL: {
            break;
        }

        case BT_DATA_SVC_DATA128: {
            //LOG_INF("Adv data");
            break;
        }
    }

    return true;
}

static void handle_received_adv_data(adv_le_user_data_t* adv_le_data)
{
    int err;
    struct bt_conn* conn;
    bt_addr_le_t *addr = peer_get_le_addr(0);

    if (connect_nok() == true) {
        //LOG_INF("Already connected. Don't create additional (non-OTA) connections");
        return;
    }

    if (bt_addr_le_is_bonded(BT_ID_DEFAULT, addr)) {
        if (get_peer_online(0)) {
            return;
        }
    }

    set_peer_rssi(adv_le_data->rssi);

    //Do not connect if already connected?
    (void)scan_stop();

    if ((err = bt_conn_le_create(addr, &create_params,
                                &conn_params,
                                &conn))) {
        LOG_ERR("Create connection failed (err %d)", err);
    } else {
        LOG_INF("Connection pendning...");
        set_client_conn(conn);
    }
}

static void new_data_thread(void *, void *, void *)
{
    adv_le_user_data_t adv_le_data;

	while (1) {
        int err = k_msgq_get(&new_data_msgq, &adv_le_data, K_FOREVER);

		if (err == 0) {
            //LOG_INF("device rssi %d", adv_le_data.rssi);

            handle_received_adv_data(&adv_le_data);
		} else {
            LOG_ERR("Get msg data returned %d", err);
        }
	}
}

static void scan_recv(const struct bt_le_scan_recv_info* info,
                      struct net_buf_simple* buf)
{
    int err;
    int idx;
    int queue_attempts = 0;
    adv_le_user_data_t adv_le_data = {.p_idx = 0, .rssi = 0, .adv_int = 0, .connected = 0, .adv_num = 0, .sensor_data = {UINT8_MAX}};

    adv_le_data.rssi = info->rssi;

    if ((idx = get_peers_idx_from_bt_addr_le_t(info->addr)) < 0) {
        return;
    }

    bt_data_parse(buf, parse_adv_data_cb, &adv_le_data);

    do {
        err = k_msgq_put(&new_data_msgq, &adv_le_data, K_NO_WAIT);

        if (err < 0) {
            LOG_ERR("Failed to add new adv data message to queue");

            k_msgq_purge(&new_data_msgq);

            queue_attempts++;
        }
    } while((err < 0) && (queue_attempts < MAX_NEW_DATA_QUEUE_ATTEMPTS));
}

static void scan_timeout(void)
{
    LOG_INF("Scan timeout");
}

static int scan_stop(void)
{
    int err;
    if ((err = bt_le_scan_stop()) == -EALREADY) {
        LOG_WRN("scan le stop already in progress");
    } else if (err) {
        LOG_ERR("Could not stop le scan (err %d)", err);
    } else {
        //LOG_INF("LE scan stopped");
    }

    return err;
}

static int scan_start(void)
{
    int err = -1;

    if ((err = bt_le_scan_start(&scan_param, NULL))) {
        LOG_ERR("Could not start le scan %d", err);
    } else {
        //LOG_INF("LE scan started");
    }

    return err;
}

int add_connectable_addr(uint8_t* addr)
{
    int res = BT_GENERAL_ERROR;

    put_addr_in_peer(addr);

    bt_dfu_smp_init(get_peer_dfu_smp(0), &dfu_init_params);

    res = SUCCESS;

    return res;
}

int get_address(uint8_t* addr_buf)
{
	int err = 0;
	bt_addr_le_t addrs[CONFIG_BT_ID_MAX];
	size_t count = sizeof(addrs);

	bt_id_get(addrs, &count);

    if (count < 1) {
        return -1;
    }

    for (size_t i = 0; i < 6; i++) {
        addr_buf[5-i] = addrs[0].a.val[i];
    }
	
	return err;
}

int ble_scanner_init(ble_cb_t ble_cb, bool settings)
{
    int err = -1;

    if ((err = bt_enable(NULL))) {
        LOG_ERR("Bluetooth init failed (err %d)", err);
        return err;
    }

    k_work_queue_init(&peer_work_queue);

    k_work_queue_start(&peer_work_queue, peer_queue_stack_area,
                   K_THREAD_STACK_SIZEOF(peer_queue_stack_area), PEER_QUEUE_PRIORITY,
                   &peer_work_queue_config);

    bt_conn_cb_register(&conn_cbs);
    bt_conn_auth_info_cb_register(&conn_auth_info_callbacks);
    bt_le_scan_cb_register(&scan_le_cb);
    scan_start();

    return err;
}
